@@ -1,6 +1,5 @@
 import json
 import requests
-
 # This is a class that is designed to include all methods and functions one needs to make an A.P.I. call
 class API(object):
 
@@ -58,8 +57,7 @@ class API(object):
                 self._response = self._response.text
 
         except Exception as e:
-            self._response['error'] = 'Internal Server Error'
-            print(e)
+            self._response['error'] = f'Internal Server Error: {e}'
         finally:
             return self._response
 
