@@ -1,4 +1,4 @@
-from typing import Dict
+from django.utils import timezone
 from bank_sync.Resources.resource import Resource
 from bank_sync.APIs.utils.generate_code import get_code
 from datetime import date, datetime
@@ -6,6 +6,10 @@ import json
 import time
 import threading
 from bank_sync.models import Callbacks
+from bank_sync.models import IPNData
+import hmac
+import hashlib
+import base64
 try:
     from django.conf import settings
 except Exception as e:
@@ -70,6 +74,9 @@ class Payments(Resource):
 
         data = {}
 
+        default_reference = payload.get("transfer", {}).get(
+            "reference", get_code(length=14))
+
         if operation is None:
             return "Specify the operation: Resource.BALANCE, Resource.MINI_STATEMENT, Resource.FULL_STATEMENT, Resource.ACCOUNT_VALIDATION or Resource.ACCOUNT_TRANSACTIONS"
 
@@ -78,7 +85,7 @@ class Payments(Resource):
             # If bank_id is COOP
             if super().get_bank_id() == super().COOP:
                 data.update({
-                    "MessageReference": f'{payload.get("transfer", {}).get("reference", get_code(length=14))}'
+                    "MessageReference": f'{default_reference}'
                 })
                 data.update({
                     "CallBackUrl": self.bank_sync_call_back.get("payments", "")
@@ -102,7 +109,7 @@ class Payments(Resource):
 
                     data["Destinations"].append(
                         {
-                            "ReferenceNumber": f'{payload.get("transfer", {}).get("reference", get_code(length=14))}_1',
+                            "ReferenceNumber": f'{default_reference}_1',
                             "AccountNumber": destinations[i].get("account_number", None),
                             "BankCode": payload.get("transfer", {}).get("bank_code", None),
                             "BranchCode": payload.get("transfer", {}).get("branch_code", None),
@@ -157,7 +164,7 @@ class Payments(Resource):
                         "BranchCode": payload.get("transfer", {}).get("branch_code", None),
                         "BeneficiaryAccountName": payload.get("destination", {}).get("name", None),
                         "Country": payload.get("transfer", {}).get("country", None),
-                        "Reference": f'{get_code(length=14)}',
+                        "Reference": f'{default_reference}',
                         "Currency": payload.get("transfer", {}).get("currency_code", None),
                         "Account": destination.get("account_number", None),
                         "Amount": destination.get("amount", None),
@@ -224,7 +231,7 @@ class Payments(Resource):
             # If bank_id is COOP
             if super().get_bank_id() == super().COOP:
                 data.update({
-                    "MessageReference": f'{get_code(length=14)}'
+                    "MessageReference": f'{default_reference}'
                 })
                 data.update({
                     "CallBackUrl": self.bank_sync_call_back.get("payments", "")
@@ -248,7 +255,7 @@ class Payments(Resource):
 
                     data["Destinations"].append(
                         {
-                            "ReferenceNumber": payload.get("transfer", {}).get("reference", None),
+                            "ReferenceNumber": f'{default_reference}_1',
                             "MobileNumber": f'{destinations[i].get("mobile_number",None)}',
                             "Amount": destinations[i].get("amount", None),
                             "Narration": payload.get("transfer", {}).get("description", "")
@@ -286,6 +293,7 @@ class Payments(Resource):
                             "reference": payload.get("transfer", {}).get("reference", None),
                             "date": payload.get("transfer", {}).get("date", None),
                             "description": payload.get("transfer", {}).get("description", ""),
+                            "callbackUrl": self.bank_sync_call_back.get("payments", ""),
                         }
                     })
             # If bank_id is DTB
@@ -328,13 +336,13 @@ class Payments(Resource):
                         "BeneficiaryAccountName": destination.get("name", None),
                         "Country": payload.get("transfer", {}).get("country", None),
                         "TranType": "Mpesa",
-                        "Reference": f'{get_code(length=14)}',
+                        "Reference": f'{default_reference}',
                         "Currency": payload.get("transfer", {}).get("currency_code", None),
                         "Account": destination.get("mobile_number", None),
                         "Amount": destination.get("amount", None),
                         "Narration": payload.get("transfer", {}).get("description", None),
                         "Transaction Date": payload.get("transfer", {}).get("date", date.today().strftime('%Y%m%d')).replace("-", ""),
-                        "Validation ID": f'{get_code(length=14)}',
+                        "Validation ID": f'{default_reference}',
                     })
         elif operation == super().RTGS:
 
@@ -389,7 +397,7 @@ class Payments(Resource):
                         "BeneficiaryAccountName": destination.get("name", None),
                         "Country": payload.get("transfer", {}).get("country", None),
                         "TranType": "RTGS",
-                        "Reference": f'{get_code(length=14)}',
+                        "Reference": f'{default_reference}',
                         "Currency": payload.get("transfer", {}).get("currency_code", None),
                         "Account": destination.get("account_number", None),
                         "Amount": destination.get("amount", None),
@@ -457,7 +465,7 @@ class Payments(Resource):
                         "BeneficiaryAccountName": destination.get("name", None),
                         "BeneficiaryName": destination.get("name", None),
                         "Country": payload.get("transfer", {}).get("country", None),
-                        "Reference": f'{get_code(length=14)}',
+                        "Reference": f'{default_reference}',
                         "Currency": payload.get("transfer", {}).get("currency_code", None),
                         "Account": destination.get("account_number", None),
                         "Amount": f'{destination.get("amount", None)}',
@@ -469,7 +477,7 @@ class Payments(Resource):
             # If bank_id is COOP
             if super().get_bank_id() == super().COOP:
                 data.update({
-                    "MessageReference": f'{get_code(length=14)}'
+                    "MessageReference": f'{default_reference}'
                 })
                 data.update({
                     "CallBackUrl": self.bank_sync_call_back.get("payments", "")
@@ -493,7 +501,7 @@ class Payments(Resource):
 
                     data["Destinations"].append(
                         {
-                            "ReferenceNumber": payload.get("transfer", {}).get("reference", None),
+                            "ReferenceNumber": f'{default_reference}_1',
                             "AccountNumber": destinations[i].get("account_number", None),
                             "BankCode": payload.get("transfer", {}).get("bank_code", None),
                             "Amount": destinations[i].get("amount", None),
@@ -549,7 +557,7 @@ class Payments(Resource):
                         "BranchCode": payload.get("transfer", {}).get("branch_code", None),
                         "BeneficiaryAccountName": destination.get("name", None),
                         "Country": payload.get("transfer", {}).get("country", None),
-                        "Reference": f'{get_code(length=14)}',
+                        "Reference": f'{default_reference}',
                         "Currency": payload.get("transfer", {}).get("currency_code", None),
                         "Account": destination.get("account_number", None),
                         "Amount": destination.get("amount", None),
@@ -561,7 +569,7 @@ class Payments(Resource):
             # If bank_id is COOP
             if super().get_bank_id() == super().COOP:
                 data.update({
-                    "MessageReference": f'{get_code(length=14)}'
+                    "MessageReference": f'{default_reference}'
                 })
                 data.update({
                     "CallBackUrl": self.bank_sync_call_back.get("payments", "")
@@ -585,7 +593,7 @@ class Payments(Resource):
 
                     data["Destinations"].append(
                         {
-                            "ReferenceNumber": payload.get("transfer", {}).get("reference", None),
+                            "ReferenceNumber": f'{default_reference}_1',
                             "PhoneNumber": destinations[i].get("mobile_number", None),
                             "Amount": destinations[i].get("amount", None),
                             "TransactionCurrency": payload.get("transfer", {}).get("currency_code", None),
@@ -727,6 +735,12 @@ class Payments(Resource):
                     })
 
         data.update(payload.get("additional_properties", {}))
+        data.update({
+            "bank_id": super().get_bank_id(),
+            "type": super().get_operation(),
+            "bank_sync_call_back": self.bank_sync_call_back.get("payments", ""),
+            "reference": default_reference
+        })
 
         return data
 
@@ -734,7 +748,6 @@ class Payments(Resource):
         return self.standardize_response(response_data=super().response())
 
     def standardize_response(self, response_data):
-        print(f'standardize_response: {response_data}')
 
         data = {
             "bank_id": super().get_bank_id(),
@@ -786,6 +799,7 @@ class Payments(Resource):
                     "messageDateTime", date.today().strftime('%d/%m/%y %H:%M:%S'))
 
         elif super().get_bank_id() == super().EQUITY:
+            print(f'standardize_response response_data: {response_data}')
             data["message"] = response_data.get("message", "")
             data["code"] = response_data.get("code", -1111111)
             if 'mobileMoneyInfo' in response_data.keys():
@@ -793,7 +807,7 @@ class Payments(Resource):
                     "mobileMoneyInfo", {}).get("ThirdPartyTranID", "")
             else:
                 data["transaction_id"] = response_data.get(
-                    "data", {}).get("transactionId", "")
+                    "data", {}).get("transactionId", response_data.get("transactionId", response_data.get("transactionId ", "")))
             data["date"] = response_data.get(
                 "date", date.today().strftime('%d/%m/%y %H:%M:%S'))
         # If bank_id is DTB
@@ -847,124 +861,227 @@ class Payments(Resource):
 
         if bool(data):
             # save the data returned to be sent back to a callback
-            self.sync_callback(response=data)
+            self.simulate_callback(response=data)
             return data
 
         return super().response()
 
-    # Use this method to standardize the callbacks sent by the payments resource
-    # get the serialised/standardized response
-    def sync_callback(self, response={}):
+    # Use this method to standardize the callbacks sent by the payments resource get the serialised/standardized response
+    # and add aditional identifiable data to it
+    # This method is used to simulate callbacks for banks or operations that return data synchronously
+    # This method is also used by banks that do not return a transaction reference in the synchronous data, otherwise this
+    # would have been handled by the core integrations
+    def simulate_callback(self, response={}):
         if super().get_bank_id() in [super().EQUITY]:
-            # get the standardized request
-            request = super().request()
-            response["bank_id"] = super().get_bank_id()
-            response["type"] = super().get_operation()
+            if super().get_operation() not in [super().MOBILE_WALLET]:
+                # get the standardized request
+                request = super().request()
+                response["bank_id"] = super().get_bank_id()
+                response["type"] = super().get_operation()
 
-            if super().get_bank_id() == super().EQUITY:
-                response["code"] = int(response.get("code"))
-                response["completed"] = False
-                response["account_reference"] = request.get(
-                    "transfer", {}).get("reference", "")
-                response["account_source"] = request.get(
-                    "source", {}).get("account_number", "")
-                response["account_destination"] = request.get(
-                    "destinations", [])[0].get("account_number", "")
-                response["amount"] = f'{request.get("destinations", [])[0].get("amount", -1)}'
+                # Equity bank retunrs data synchronously, we will append to it unique identifiers that are not returned
+                # e.g. our unique bank id and type of operation/transaction performed
+                if super().get_bank_id() == super().EQUITY:
+                    response["code"] = int(response.get("code"))
+                    response["completed"] = False
+                    response["account_reference"] = request.get(
+                        "transfer", {}).get("reference", "")
+                    response["account_source"] = request.get(
+                        "source", {}).get("account_number", "")
+                    response["account_destination"] = request.get(
+                        "destinations", [])[0].get("account_number", "")
+                    response["amount"] = f'{request.get("destinations", [])[0].get("amount", -1)}'
 
-                if response.get("code") == 0:
-                    response["completed"] = True
+                    if response.get("code") == 0:
+                        response["completed"] = True
 
-                callback_url = request.get(
-                    "transfer", {}).get("callback_url", "")
+                    callback_url = request.get(
+                        "transfer", {}).get("callback_url", "")
 
-                t1 = threading.Thread(target=self._send_to_callback,
-                                      args=[response, callback_url])
-                t1.start()
+                    # simultate a callback by sending data later via a thread
+                    # this will only work if a callback was sent/placed in the standardized payload
+                    threading.Thread(target=self._send_to_callback,
+                                     args=[response, callback_url]).start()
 
-    def async_callback(self, callback_data={}, forward=True):
-        print(f'callback_data: {callback_data}')
+    # This method is used to standardize callbacks sent
+    def standardize_callback(self, callback_data={}, forward=True):
 
-        response = {"completed": False}
+        response = {
+            "completed": False,
+            "type": callback_data.get("type", -1),
+            "bank_id": callback_data.get("bank_id", -1)
+        }
 
-        if 'MessageReference' in callback_data.keys() or 'messageReference' in callback_data.keys():
-            response["bank_id"] = super().COOP
+        # Use the reference of the callback sent to get the callback data that was linked to the operation
+        # This will returns a queryset, you will use .first() method or len() to manipultate it
+        saved_callback = Callbacks.objects.filter(
+            reference=callback_data.get("reference", ""))
 
-            callback = Callbacks.objects.filter(
-                reference=callback_data.get("MessageReference", callback_data.get("messageReference", "")))
+        self.save_callback_data(saved_callback, callback_data)
 
-            if len(callback) == 0:
-                if len(callback_data.get("Destinations", [])):
-                    callback = Callbacks.objects.filter(reference=callback_data.get(
-                        "Destinations", [])[0].get("ReferenceNumber", ""))
+        callback = None
 
-            if len(callback):
-                callback = callback.first()
-                response["type"] = callback.type_code
-                response["code"] = int(callback_data.get(
-                    "MessageCode", callback_data.get("messageCode", -1)))
-                response["account_reference"] = callback_data.get(
-                    "MessageReference", callback_data.get("messageReference", ""))
-                response["date"] = callback_data.get(
-                    "MessageDateTime", callback_data.get("messageDateTime", ""))
+        # The bank id and type sent here, will be used to know how to standardize the data received
 
-                if 'Source' in callback_data.keys():
-                    response["account_source"] = callback_data.get(
-                        "Source", {}).get("AccountNumber", "")
-                elif 'source' in callback_data.keys():
-                    response["account_source"] = callback_data.get(
-                        "source", {}).get("accountNumber", "")
+        if callback_data.get("bank_id", -1) == super().COOP:
 
-                if 'Source' in callback_data.keys():
-                    response["message"] = callback_data.get("MessageDescription", "")
-                elif 'source' in callback_data.keys():
-                    response["message"] = callback_data.get(
-                        "source", {}).get("narration", "")
+            if callback_data.get("type", -1) in [super().IFT, super().MOBILE_WALLET, super().PESALINK_TO_BANK, super().PESALINK_TO_MOBILE]:
 
-                if 'Destinations' in callback_data.keys():
-                    if isinstance(callback_data.get("Destinations", {}),dict):
-                        response["account_destination"] = callback_data.get(
-                            "Destinations", {}).get("AccountNumber", "")
-                    if isinstance(callback_data.get("Destinations", []),list):
-                        response["account_destination"] = callback_data.get(
-                            "Destinations", [])[0].get("AccountNumber", "")
-                        
-                        if callback.type_code == super().MOBILE_WALLET:
+                if len(saved_callback) == 0:
+                    if len(callback_data.get("Destinations", [])):
+                        callback = Callbacks.objects.filter(reference=callback_data.get(
+                            "Destinations", [])[0].get("ReferenceNumber", ""))
+
+                if len(saved_callback):
+                    callback = saved_callback.first()
+
+                    response["code"] = int(callback_data.get(
+                        "MessageCode", callback_data.get("messageCode", -1)))
+                    response["account_reference"] = callback_data.get(
+                        "reference", "")
+                    response["date"] = callback_data.get(
+                        "MessageDateTime", callback_data.get("messageDateTime", ""))
+
+                    if 'Source' in callback_data.keys():
+                        response["account_source"] = callback_data.get(
+                            "Source", {}).get("AccountNumber", "")
+                    elif 'source' in callback_data.keys():
+                        response["account_source"] = callback_data.get(
+                            "source", {}).get("accountNumber", "")
+
+                    if 'Source' in callback_data.keys():
+                        response["message"] = callback_data.get(
+                            "MessageDescription", "")
+                    elif 'source' in callback_data.keys():
+                        response["message"] = callback_data.get(
+                            "source", {}).get("narration", "")
+
+                    if 'Destinations' in callback_data.keys():
+                        if isinstance(callback_data.get("Destinations", {}), dict):
                             response["account_destination"] = callback_data.get(
-                            "Destinations", [])[0].get("MobileNumber", "")
+                                "Destinations", {}).get("AccountNumber", "")
+                        if isinstance(callback_data.get("Destinations", []), list):
+                            response["account_destination"] = callback_data.get(
+                                "Destinations", [])[0].get("AccountNumber", "")
 
-                elif 'destination' in callback_data.keys():
-                    response["account_destination"] = callback_data.get(
-                        "destination", {}).get("accountNumber", "")
+                            if callback.type_code == super().MOBILE_WALLET:
+                                response["account_destination"] = callback_data.get(
+                                    "Destinations", [])[0].get("MobileNumber", "")
 
-                if 'Destinations' in callback_data.keys():                    
-                    if isinstance(callback_data.get("Destinations", {}),dict):
+                    elif 'destination' in callback_data.keys():
+                        response["account_destination"] = callback_data.get(
+                            "destination", {}).get("accountNumber", "")
+
+                    if 'Destinations' in callback_data.keys():
+                        if isinstance(callback_data.get("Destinations", {}), dict):
+                            response["transaction_id"] = callback_data.get(
+                                "Destinations", {}).get("TransactionID", "")
+                        if isinstance(callback_data.get("Destinations", []), list):
+                            response["transaction_id"] = callback_data.get(
+                                "Destinations", [])[0].get("TransactionID", "")
+                    elif 'destination' in callback_data.keys():
                         response["transaction_id"] = callback_data.get(
-                            "Destinations", {}).get("TransactionID", "")
-                    if isinstance(callback_data.get("Destinations", []),list):
-                        response["transaction_id"] = callback_data.get(
-                            "Destinations", [])[0].get("TransactionID", "")
-                elif 'destination' in callback_data.keys():
-                    response["transaction_id"] = callback_data.get(
-                        "destination", {}).get("transactionID", "")
+                            "destination", {}).get("transactionID", "")
 
-                if 'Destinations' in callback_data.keys():                                     
-                    if isinstance(callback_data.get("Destinations", {}),dict):
-                        response["amount"] = f'{callback_data.get("Destinations", {}).get("Amount", "")}' 
-                    if isinstance(callback_data.get("Destinations", []),list):
-                        response["amount"] = f'{callback_data.get("Destinations", [])[0].get("Amount", "")}'
-                elif 'destination' in callback_data.keys():
-                    response["amount"] = f'{callback_data.get("destination", {}).get("amount", "")}'
+                    if 'Destinations' in callback_data.keys():
+                        if isinstance(callback_data.get("Destinations", {}), dict):
+                            response["amount"] = f'{callback_data.get("Destinations", {}).get("Amount", "")}'
+                        if isinstance(callback_data.get("Destinations", []), list):
+                            response["amount"] = f'{callback_data.get("Destinations", [])[0].get("Amount", "")}'
+                    elif 'destination' in callback_data.keys():
+                        response["amount"] = f'{callback_data.get("destination", {}).get("amount", "")}'
 
-                if response.get("code") == 0:
-                    response["completed"] = True
+                    if response.get("code") == 0:
+                        response["completed"] = True
 
-                if forward:
-                    self._send_to_callback(
-                        response, callback_url=callback.callback, sleep=False)
+                    callback.response = callback_data
+                    callback.save()
+
+        elif callback_data.get("bank_id", -1) == super().EQUITY:
+
+            if callback_data.get("type", -1) in [super().MOBILE_WALLET]:
+
+                if len(saved_callback):
+                    callback = saved_callback.first()
+                    response.update({
+                        "message": f'{callback_data.get("message", "")}: {callback_data.get("data", {}).get("ResponseDescription","")}',
+                        "code": round(float(callback_data.get("code", -1)), 2),
+                        "transaction_id": callback_data.get("data", {}).get("ThirdPartyTranID", ""),
+                        "date": callback.request.get("transfer", {}).get('date', ""),
+                        "account_reference": callback_data.get("transactionReference", ""),
+                        "account_source": f"{callback.request.get('source',{}).get('account_number','')} {callback.request.get('source',{}).get('name','')}",
+                        "account_destination": callback_data.get("data", {}).get("ReceiverMsisdn", ""),
+                        "amount": round(float(callback.request.get("source", {}).get('amount', -1)), 2)
+                    })
+
+                    if response.get("code") == 0:
+                        response["completed"] = True
+
+        # This is a variable used to forward the standardized callback data to another callback url
+        # If it was supplied ans saved, else return the standardized response back to the
+        if forward:
+            if callback is not None:
+                self._send_to_callback(
+                    response, callback_url=callback.callback, sleep=False)
 
         return response
 
+    # This method is used to standardize callbacks sent
+    def standardize_ipn(self, ipn_data={}, forward=True):
+
+        ipn = None
+        
+        response = {
+            "completed": False,
+            "type": ipn_data.get("type", -1),
+            "bank_id": ipn_data.get("bank_id", -1)
+        }
+
+        # Use the reference of the callback sent to get the callback data that was linked to the operation
+        # This will returns a queryset, you will use .first() method or len() to manipultate it
+        saved_ipn = super().get_ipn_object.filter(
+            client_id=ipn_data.get("client_id", ""), bank_id=ipn_data.get("bank_id", ""), account_number=ipn_data.get("bank", {}).get("account", ""))
+
+        if len(saved_ipn):
+            # get the IPN
+            ipn = saved_ipn.first()
+
+            # the IPN comes back with a signature, regenerate this signature using the saved client secret as the key
+            # and by concatinating the client_id:bank_account_number
+            if ipn_data.get("signature", "") == self.generate_signature(secret=ipn.client_secret, message=f'{ipn_data.get("client_id", "")}:{ipn_data.get("bank",{}).get("account","")}'):
+                
+                # save the response sent back to the ipn
+                self.save_ipn_data(saved_ipn, ipn_data)
+
+                if ipn_data.get("bank_id", -1) == super().EQUITY:
+
+                    if ipn_data.get("type", -1) in [super().OMN]:
+
+                        response.update({
+                            "message": f'{ipn_data.get("transaction", {}).get("status","")} {ipn_data.get("transaction", {}).get("remarks", "")}',
+                            "code": ipn_data.get("code", -1),
+                            "transaction_id": ipn_data.get("transaction", {}).get("reference", ""),
+                            "date": ipn_data.get("transaction", {}).get('date', ""),
+                            "account_reference": f'{ipn_data.get("bank", {}).get("reference", "")}',
+                            "account_source": f"{ipn_data.get('bank',{}).get('account','')}",
+                            "account_destination": ipn_data.get("transaction", {}).get("remarks", ""),
+                            "amount": ipn_data.get("transaction", {}).get('amount', -1)
+                        })
+
+                        if response.get("code") == 0:
+                            response["completed"] = True
+
+                # This is a variable used to forward the standardized ipn data to another callback url
+                # If it was supplied and saved, else return the standardized response back to the
+                if forward:
+                    if ipn is not None:
+                        self._send_to_callback(
+                            response, callback_url=ipn.url, sleep=False)
+
+        return response
+
+    # Use this method to send data to a callback
+
     def _send_to_callback(self, response, callback_url='', sleep=True, sleep_time=1):
         if sleep:
             time.sleep(1*sleep_time)
@@ -976,10 +1093,37 @@ class Payments(Resource):
             self.api_request(payload=payload,
                              method='POST', verify=True)
 
+    # Use this method to set an operation/transaction to save/register callbacks
+    # When an operation is called with this method, it saves the callback url sent and the reference
+    # This method regesiters callbacks for banks that return data as a callback with the reference sent
     def register_callback(self):
-        if super().get_bank_id() in [super().COOP]:
+        save_callback = True
+        if super().get_bank_id() in [super().COOP, super().EQUITY]:
+            if super().EQUITY == super().get_bank_id() and super().get_operation() not in [super().MOBILE_WALLET]:
+                save_callback = False
+
+        if save_callback:
+            # Get the request that was sent
             request = super().request()
+            # save the reference, callback, bank id and operation
             Callbacks.objects.create(reference=request.get("transfer", {}).get("reference", ""),
                                      callback=request.get("transfer", {}).get(
                                          "callback_url", ""),
                                      request=request, bank_id=super().get_bank_id(), type_code=super().get_operation())
+
+    def save_callback_data(self, saved_callback, callback_data):
+        if len(saved_callback):
+            callback = saved_callback.first()
+            callback.response = callback_data
+            callback.response_time = timezone.now()
+            callback.save()
+
+    def save_ipn_data(self, saved_ipn, ipn_data):
+        if len(saved_ipn):
+            saved_ipn = saved_ipn.first()
+            IPNData.objects.create(ipn=saved_ipn, response=ipn_data)
+
+    def generate_signature(self, secret, message):
+        digest = hmac.new(secret.encode(), msg=message.encode(
+        ), digestmod=hashlib.sha256).digest()
+        return base64.b64encode(digest).decode()
